#!/usr/bin/env python
import argparse
import itertools
import os
import shutil
import logging
import sys
import multiprocessing
from multiprocessing.pool import ThreadPool

from ruv_dl.programs import ProgramFetcher
from ruv_dl.crawler import Crawler
from ruv_dl.downloader import Downloader
from ruv_dl.constants import DEFAULT_VIDEO_DESTINATION, CACHE_LOCATION


logger = logging.getLogger('ruv_dl')
handler = logging.StreamHandler(sys.stdout)
format_str = '%(asctime)s - %(name)s -  %(message)s'
date_format = '%Y-%m-%d %H:%M:%S'
formatter = logging.Formatter(format_str, date_format)
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.WARN)


def run(args):
    fetcher = ProgramFetcher(args.query, args.update, args.destination)
    with ThreadPool(8) as pool:
        programs = {}
        for program in fetcher.get_programs():
            logger.info(f'------ {program["title"]} [{program["id"]}] ------')
            crawler = Crawler(
                days_between_episodes=args.days_between_episodes,
                iteration_count=args.iteration_count,
                program=program,
            )
            programs[program['id']] = {
                'program': program,
                'episodes': pool.apply_async(
                    crawler.search_for_episodes
                )
            }

        downloaders = []
        for program_id, data in programs.items():
            downloader = Downloader(
                destination=args.destination,
                program=data['program'],
                episode_entries=data['episodes'].get(),
                threaded=not args.sequential,
            )
            entries = downloader.organize()
            downloaders.append((downloader, entries))

    total_entries_to_download = sum(
        len(entries) for _, entries in downloaders
    )
    if not total_entries_to_download:
        logger.info('No entries to download, bye')
    else:
        logger.warning(f'Downloading {total_entries_to_download} files...')
        if args.dryrun:
            logger.warning('Dryrun, not downloading anything, bye')
            return
        if args.sequential:
            results = [
                [
                    downloader.download_file(entry)
                    for entry in entries
                ] for downloader, entries in downloaders
            ]
        else:
            with ThreadPool(8) as pool:
                pool.__exit__
                async_results = [
                    pool.map_async(downloader.download_file, entries)
                    for downloader, entries in downloaders
                ]
                results = [result.get() for result in async_results]
        logger.warning(
            f'{len([r for r in itertools.chain(*results) if r])} '
            'files downloaded'
        )


def main():
    parser = argparse.ArgumentParser()
    query_arg = parser.add_argument(
        'query',
        help='Search terms to search for programs.',
        nargs='*',
    )
    parser.add_argument(
        '-u', '--update',
        action='store_true',
        help='Search for all saved shows in `--destination` and download '
        'available episodes'
    )
    parser.add_argument(
        '--destination', default=DEFAULT_VIDEO_DESTINATION, nargs='?',
        type=os.path.abspath,
        help='Top level destination directory.'
    )
    parser.add_argument(
        '--days-between-episodes', type=int, default=7, nargs='?',
        help='Rate of episode release.'
    )
    parser.add_argument(
        '--iteration-count', type=int, default=5, nargs='?',
        help='Maximum days to allow for now shows found.'
    )
    parser.add_argument(
        '--empty-cache',
        action='store_true',
        help='Empty request cache to api.ruv.is before running.'
    )
    parser.add_argument(
        '--sequential',
        action='store_true',
        help='Do not run threaded, only download one file at a time.'
    )
    parser.add_argument(
        '--dryrun',
        action='store_true',
        help='Only search and organize episodes, do not download them.'
    )
    parser.add_argument(
        '-v', '--verbosity', action='count',
        help='Increase output verbosity'
    )
    args = parser.parse_args()
    if bool(args.query) == bool(args.update):
        raise argparse.ArgumentError(
            query_arg,
            'Query terms and update are mutually exclusive and either must '
            'be included'
        )
    if args.verbosity is not None:
        multiprocessing_logger = multiprocessing.get_logger()
        if args.verbosity > 2:
            multiprocessing_logger.addHandler(handler)
            multiprocessing_logger.setLevel(logging.DEBUG)
        if args.verbosity > 1:
            logger.setLevel(logging.DEBUG)
            multiprocessing_logger.addHandler(handler)
            multiprocessing_logger.setLevel(logging.INFO)
        elif args.verbosity > 0:
            logger.setLevel(logging.INFO)
    if args.empty_cache:
        if os.path.exists(CACHE_LOCATION):
            shutil.rmtree(CACHE_LOCATION)
    os.makedirs(args.destination, exist_ok=True)
    os.makedirs(CACHE_LOCATION, exist_ok=True)
    run(args)



#!/usr/bin/env python
import json
import os
import logging
import glob

import requests

from ruv_dl.constants import PROGRAM_INFO_FN

logger = logging.getLogger(__name__)


class ProgramFetcher:
    pool = None

    def __init__(self, query, update, destination):
        self.query = query
        self.update = update
        self.destination = destination

    def get_programs(self):
        if self.query:
            return self.get_programs_by_query(self.query)
        return self.get_programs_to_update()

    def get_programs_by_query(self, queries):
        for query in queries:
            if query.isdigit():
                program_id = query
            else:
                program_id = self.get_program_id(query)
            r = requests.get(
                f'https://api.ruv.is/api/programs/program/{program_id}/all'
            )
            if r.ok:
                yield r.json()
            else:
                logger.warning(
                    f'Request for program {program_id} (query {query}) '
                    f'failed with status code {r.status_code}.'
                )

    def get_program_id(self, query):
        r = requests.get(
            f'https://api.ruv.is/api/programs/search/tv/{query}'
        )
        r.raise_for_status()
        programs = r.json()['programs']
        if not programs:
            raise RuntimeError(f'No programs found matching {query}')
        while True:
            for i, program in enumerate(programs):
                print(i + 1, ':', program['title'])
            selection = input('Select program: ')
            if selection.isdigit():
                selection = int(selection)
                if selection > 0 and selection <= len(programs):
                    return programs[selection - 1]['id']

    def get_programs_to_update(self):
        for program_info in glob.glob(
            os.path.join(self.destination, '*', PROGRAM_INFO_FN)
        ):
            with open(program_info, 'r') as f:
                try:
                    data = json.loads(f.read())
                except ValueError:
                    logger.info(f'Could not parse {program_info}')
                if 'program' in data:
                    if 'id' in data['program']:
                        yield data['program']
                    else:
                        logger.info(
                            f'Could not get program id from {data["program"]}'
                        )
                else:
                    logger.info(f'Could not get program from {data}')



#!/usr/bin/env python
import os
import datetime

from ruv_dl.constants import DATE_FORMAT


class Entry:
    def __init__(self, fn, url, date, etag):
        self.fn = fn
        self.url = url
        self.date = date
        self.etag = etag
        self.target_path = None

    def to_dict(self):
        return {
            'fn': self.fn,
            'url': self.url,
            'date': self.date.strftime(DATE_FORMAT),
            'etag': self.etag,
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            fn=data['fn'],
            url=data['url'],
            date=datetime.datetime.strptime(data['date'], DATE_FORMAT),
            etag=data['etag'],
        )

    def set_target_path(self, path):
        self.target_path = path

    def exists_on_disk(self):
        if self.target_path is None:
            raise RuntimeError(f'Missing target path for {self.to_dict}')
        return os.path.exists(self.target_path)

    def __hash__(self):
        return hash(self.etag)
        # return hash((self.fn, self.url, self.date, self.etag))

    def __eq__(self, other):
        return isinstance(other, Entry) and hash(self) == hash(other)
